/*******************************************************************************
/
/  MessageBoxInstrument.C
/
/  DESCRIPTION:
/     Provides sample code for the interception and handling of windows
/   displayed on a desktop in a headless environment.
/
/     This code is designed to run as a Win32 service.  It schedules itself via
/   a waitable timer to wake up at a specified interval, enumerate windows
/   displayed on the desktop, and handle those windows of interest; in this
/   case, certain NET SEND messages that are sent to the computer remotely.
/
/     Depending upon your specific application, you may want to have the
/   service constantly running, or, perhaps, not make it a service at all. The
/   code controlling scheduling and service dispatch is provided as a suggestion
/   and convenience to the embedded developer.
/
/  Copyright (c) 2000, Microsoft Corporation
/
*******************************************************************************/
#include "MessageBoxInstrument.h"



//
//  GLOBALS
//

//
// This is the table used by the SCM to determine the entry point or entry
//  points defined for each service contained within the compiled executable.
// In this case, our binary will only provide one service, entered through
//  ServiceMain().
//
SERVICE_TABLE_ENTRY     g_steTable[ 2 ] = 
                          { { SERVICE_NAME, ServiceMain },
                            { NULL,         NULL        } };

//
// Handle to our service status, returned by RegisterServiceCtrlHandler().
//
SERVICE_STATUS_HANDLE   g_hSvcStatus = (SERVICE_STATUS_HANDLE)NULL;

//
// Global SERVICE_STATUS structure.  Used to update the SCM.
//
SERVICE_STATUS          g_svcStatus = { 0 };

//
// Handle to the window enumeration thread.
//
HANDLE                  g_hEnumThread = NULL;

BOOL    g_bConsole;
HANDLE  g_hConIn;
HANDLE  g_hConOut;
TCHAR   g_szOutputString[ 1024 ];
DWORD   g_dwWritten;


// internal function prototypes
VOID CmdInstallService();
VOID CmdRemoveService();
LPTSTR GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize );

//
//  CODE
//


//+++**************************************************************************
//  main()
//
//  DESCRIPTION:
//    Main entry point for process.  Starts the service control dispatcher,
//      informing the SCM that the service entry point is ServiceMain().
//
//  PARAMS:
//    int   argc    - number of arguments in argv (not used)
//    char  *argv[] - pointer to string array containing arguments (not used)
//
//  RETURNS:
//    0
//
//---**************************************************************************
int
main(
  int   argc,
  char  *argv[] 
  )
{
  
	
   if ( (argc > 1) &&
        ((*argv[1] == '-') || (*argv[1] == '/')) )
   {
      if ( _stricmp( "install", argv[1]+1 ) == 0 )
      {
         CmdInstallService();
      }
      else if ( _stricmp( "remove", argv[1]+1 ) == 0 )
      {
         CmdRemoveService();
      }
      else
      {
         goto dispatch;
      }
      exit(0);
   }

   // if it doesn't match any of the above parameters
   // the service control manager may be starting the service
   // so we must call StartServiceCtrlDispatcher
   dispatch:
   // this is just to be friendly
   printf( "%s -install          to install the service\n", SZAPPNAME );
   printf( "%s -remove           to remove the service\n", SZAPPNAME );
   printf( "%s -debug <params>   to run as a console app for debugging\n", SZAPPNAME );
   printf( "\nStartServiceCtrlDispatcher being called.\n" );
   printf( "This may take several seconds.  Please wait.\n" );

	
  //
  //  Register the service control dispatcher.  The SCM will, in turn, call
  //    the ServiceMain() function.
  //
  StartServiceCtrlDispatcher( g_steTable );

  return 0;
}


//+++**************************************************************************
//  ServiceMain()
//
//  DESCRIPTION:
//    Main entry point for service (called by Service Control Manager).
//      Registers the service control handler (Handler()) with the SCM.  Updates
//      status, performs initialization, waits for the main thread to terminate,
//      then returns.  ServiceMain() is not to return until the service is
//      ready to terminate.
//
//  PARAMS:
//    DWORD   dwArgc    - number of arguments in lpszArgv (not used)
//    LPTSTR  *lpszArgv - pointer to string array containing arguments 
//                        (not used)
//
//  RETURNS:
//    VOID
//
//---**************************************************************************
VOID
WINAPI
ServiceMain(
  DWORD   dwArgc,
  LPTSTR  *lpszArgv 
  )
{
  TCHAR   szErr[ 256 ] = { 0 };
  //
  //  Register the control callback
  //
  g_hSvcStatus = RegisterServiceCtrlHandler( SERVICE_NAME, Handler );

  if( !g_hSvcStatus )
  {
    //
    //  Add structured error handling
    //
    return;
  }

  //
  //  Initialize the debug console.
  //
  SOM_INIT();
  INIT_DBG_CONSOLE();
  //
  //  Let the SCM know what our status is.  We are now up and running, within
  //    our own process, and we accept SERVICE_CONTROL_STOP and
  //    SERVICE_CONTROL_SHUTDOWN control codes from the SCM.
  //
  g_svcStatus.dwCurrentState = SERVICE_RUNNING;
  g_svcStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  g_svcStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
                                   SERVICE_ACCEPT_SHUTDOWN;

  SetServiceStatus( g_hSvcStatus, &g_svcStatus );

  //
  //  Perform any aditional initialization required here.  All we need to do is
  //    create a window enumeration thread.
  //
  g_hEnumThread = CreateThread( NULL,       // inherit default SD
                                0,          // inherit current stack size
                                ThreadProc, // use ThreadProc thread procedure
                                NULL,       // no parameter required
                                0,          // use default flags
                                NULL );     // we don't need the TID

  if( !g_hEnumThread )
  {
    DEBUG_OUT( TEXT( "Could not create enumeration thread..." ) );
    //
    //  Add error handling.
    //
    return;
  }

  while( TRUE )
  {
    WaitForSingleObject( g_hEnumThread, 5000 );
    Sleep( 5000 );
  }

  //
  //  Add error handling here in case of WAIT_FAILED
  //

  FREE_DBG_CONSOLE(FALSE);
  CloseHandle( g_hEnumThread );
  
  return;
}


//+++**************************************************************************
//  Handler()
//
//  DESCRIPTION:
//    Control dispatch for service.  Receives and handles all start, stop, and
//      shutdown requests, etc.
//
//  PARAMS:
//    DWORD   fdwControl -  Service control flags
//
//  RETURNS:
//    VOID
//
//---**************************************************************************
VOID
WINAPI
Handler(
  DWORD   fdwControl 
  )
{
  TCHAR           szErr[ 256 ] = { 0 };
  static HANDLE   hStopEvent = NULL;
  DWORD           dwWaitRet = 0;
  
  switch( fdwControl )
  {
    case SERVICE_CONTROL_SHUTDOWN:
    case SERVICE_CONTROL_STOP:
    {
      //
      // Open the enumeration thread's stop event.  If, for some reason, we
      //  cannot open the event, set the event, or wait on the thread, then fall
      //  though, as there is little more we can do.
      //
      hStopEvent = OpenEvent( EVENT_ALL_ACCESS,
                              FALSE,
                              EVENT_NAME );

      if( hStopEvent && g_hEnumThread )
      {
        SetEvent( hStopEvent );

        dwWaitRet = WaitForSingleObject( g_hEnumThread, THREAD_TIMEOUT );

        switch( dwWaitRet )
        {
          case WAIT_OBJECT_0:
          {
            g_svcStatus.dwWin32ExitCode = ERROR_SUCCESS;
            
            CloseHandle( hStopEvent );
            hStopEvent = NULL;

            CloseHandle( g_hEnumThread );
            g_hEnumThread = NULL;

            break;
          }
          case WAIT_FAILED:
          case WAIT_TIMEOUT:
          default:
          {
            DEBUG_OUT( TEXT( "WaitForSingleObject returned WAIT_FAILED, " )
                       TEXT( "WAIT_TIMEOUT or an invalid value" ) );
            //
            // Add structured error handling here.
            //
            g_svcStatus.dwWin32ExitCode = GetLastError();
            
            CloseHandle( hStopEvent );
            hStopEvent = NULL;

            CloseHandle( g_hEnumThread );
            g_hEnumThread = NULL;

            break;
          }
        }
      }
      else
      {
        TCHAR szErr [ 256 ] = { 0 };

        //
        // Add structured error handling here.
        //
        if( g_hEnumThread )
        {
          CloseHandle( g_hEnumThread );
          g_hEnumThread = NULL;
        }

        if( hStopEvent )
        {
          CloseHandle( hStopEvent );
          hStopEvent = NULL;
        }

        g_svcStatus.dwWin32ExitCode = ERROR_INVALID_DATA;
      }
      
      g_svcStatus.dwCurrentState = SERVICE_STOPPED;
      break;
    }
    case SERVICE_CONTROL_INTERROGATE:
    {
      g_svcStatus.dwWin32ExitCode = ERROR_SUCCESS;
      g_svcStatus.dwCurrentState = SERVICE_RUNNING;
      break;
    }
    default:
    {
      g_svcStatus.dwWin32ExitCode = ERROR_INVALID_PARAMETER;
      break;
    }
  }

  SetServiceStatus( g_hSvcStatus, &g_svcStatus );
  
  return;
}


//+++**************************************************************************
//  ThreadProc()
//
//  DESCRIPTION:
//    Thread procedure for main "watch thread".  Schedules itself to enumerate
//      the windows on the desktop once per second via a waitable timer.  Also
//      waits for a stop event from the Handler() function.
//
//  PARAMS:
//    LPVOID  lpvParam - pointer to a function-defined parameter (not used)
//
//  RETURNS:
//    ERROR_SUCCESS on success
//    A Win32 error code on error
//    -1 on unknown error
//
//---**************************************************************************
DWORD
WINAPI
ThreadProc(
  LPVOID    lpvParam 
  )
{
  DWORD       rc = ERROR_SUCCESS;
  DWORD       dwWaitRet = 0;
  HANDLE      hWaitableTimer = NULL;
  HANDLE      hStopEvent = NULL;
  HANDLE      hDesktop = NULL;
  FILETIME    ft = { 0 };
  SYSTEMTIME  st = { 0 };
  LARGE_INTEGER
              li = { 0 };
  HANDLE      rghWaitArray[ 2 ];
    TCHAR szErr[256]={0};

  DEBUG_OUT( TEXT( "Thread started..." ) );
  
  hStopEvent = CreateEvent( NULL,
                            FALSE,
                            FALSE,
                            EVENT_NAME );

  if( !hStopEvent )
  {
    _stprintf( szErr, TEXT( "%d" ), GetLastError() );
    DEBUG_OUT( TEXT( "ThreadProc: !hStopEvent" ) );
    DEBUG_OUT(szErr );
  }
  else
  {
    DEBUG_OUT( TEXT( "ThreadProc: Event created" ) );
  }
  hWaitableTimer = CreateWaitableTimer( NULL,  // inherit default SD
                                        FALSE, // syncronization timer
                                        NULL );// unnamed

  while( TRUE )
  {
    GetSystemTime( &st );

    _stprintf( szErr, 
                TEXT( "Sched: %2.2d/%2.2d/%4.4d %2.2d:%2.2d:%2.2d.%3.3d UTC" ),
                st.wMonth, st.wDay, st.wYear, st.wHour, st.wMinute, st.wSecond,
                st.wMilliseconds );
                
    DEBUG_OUT( szErr );
    
    SystemTimeToFileTime( &st, &ft );
    
    memcpy( &li, &ft, sizeof( ft ) );
    
    li.QuadPart += FILETIMES_IN_SECOND;
    
    memcpy( &ft, &li, sizeof( li ) );
    
    FileTimeToSystemTime( &ft, &st );
    
    _stprintf( szErr, 
                TEXT( "Sched: %2.2d/%2.2d/%4.4d %2.2d:%2.2d:%2.2d.%3.3d UTC" ),
                st.wMonth, st.wDay, st.wYear, st.wHour, st.wMinute, st.wSecond,
                st.wMilliseconds );

    DEBUG_OUT( szErr );
    
    if( !SetWaitableTimer( hWaitableTimer,
                           &li,
                           0,
                           NULL,
                           NULL,
                           FALSE ) )
    {
      rc = GetLastError();
      _stprintf( szErr, TEXT( "%d" ), rc );
      DEBUG_OUT( TEXT( "ThreadProc: Could not set waitable timer" ) );
      DEBUG_OUT( szErr );
      goto EXIT;
    }

    rghWaitArray[ 0 ] = hStopEvent;
    rghWaitArray[ 1 ] = hWaitableTimer;
    
    dwWaitRet = 
      WaitForMultipleObjectsEx( sizeof( rghWaitArray ) / sizeof( HANDLE ),
                                rghWaitArray, 
                                FALSE,
                                FILETIMES_IN_SECOND * 20, // wait 20 sec max
                                TRUE );

    switch( dwWaitRet )
    {
      case WAIT_TIMEOUT:
      {
        DEBUG_OUT( TEXT( "ThreadProc: WAIT_TIMEOUT" ) );
        //
        // The waitable timer is dead for some reason.
        //  Add structured error handling.
        //
        rc = WAIT_TIMEOUT;
        goto EXIT;
      }
      case WAIT_FAILED:
      {
        DEBUG_OUT( TEXT( "ThreadProc: WAIT_FAILED" ) );
        //
        //  Add structured error handling.
        //
        rc = GetLastError();
        goto EXIT;
      }
      case WAIT_OBJECT_0:
      {
        goto EXIT;
      }
      case WAIT_OBJECT_0 + 1:
      {
        RunEnumeration();
        CancelWaitableTimer( hWaitableTimer );
        continue;
      }
      default:
      {
        //
        //  Add structured error handling.
        //
        rc = -1;
        goto EXIT;
      }
    }
  }
  
EXIT:

  CancelWaitableTimer( hWaitableTimer );
  
  if( rc != ERROR_SUCCESS )
  {
    _stprintf( szErr, TEXT( "%d" ), GetLastError() );
    DEBUG_OUT( TEXT( "Exiting thread" ) );
    DEBUG_OUT( szErr );
  }
  
  if( hWaitableTimer )
  {
    CancelWaitableTimer( hWaitableTimer );
    CloseHandle( hWaitableTimer );
  }

  if( hStopEvent )
  {
    CloseHandle( hStopEvent );
  }

  return rc;
}


//+++**************************************************************************
//  RunEnumeration()
//
//  DESCRIPTION:
//    Run whenever the waitable timer signals.  Starts window enumeration.  If
//      the window is found, completes application-specific handling.
//
//  PARAMS:
//    VOID
//
//  RETURNS:
//    VOID
//
//---**************************************************************************
VOID
CALLBACK
RunEnumeration(
  VOID
  )
{
  DWORD       rc = 0;
  HWND        hTargetWindow = NULL;

  EnumDesktopWindows( GetThreadDesktop(GetCurrentThreadId()),
                            EnumWindowsProc,
                            (LPARAM)&hTargetWindow );
                            
  if( NULL != hTargetWindow )
  {
    //
    //  We found the window.  Do whatever is appropriate.
    //
    DoApplicationSpecificHandling( hTargetWindow );
  }

}


//+++**************************************************************************
//  EnumWindowsProc()
//
//  DESCRIPTION:
//    A WNDENUMPROC-type function.  Called for each top-level window on the
//      desktop.  Decides if the top-level window is of the type we're looking
//      for.  If so, enumerates through the child windows to find specific text
//      within a static control.  If the window is determined to be the exact
//      window in question, the window handle is returned to the main thread via
//      lParam.
//
//  PARAMS:
//    HWND    hWnd    - handle to current top-level window
//    LPARAM  lParam  - points to an HWND to recieve the handle of the window
//                      in question [IN / OUT]
//
//  RETURNS:
//    FALSE (stops enumeration) on fatal error or if the window is found
//    TRUE (continues enumeration) if the supplied handle is not the window
//      we are looking for
//
//---**************************************************************************
BOOL
CALLBACK
EnumWindowsProc(
  HWND    hWnd,
  LPARAM  lParam 
  )
{
  BOOL    rc = TRUE;
  TCHAR   szBuffer[ 256 ] = { 0 };
  HWND    *phTargetWindow = (HWND*)lParam;
  BOOL    bTextFound = FALSE;

  if( !lParam )
  {
    rc = FALSE;
    goto EXIT;
  }

  //
  //  First, see if the window has the title text we are looking for.
  //
  if( GetWindowText( hWnd, szBuffer, sizeof( szBuffer ) ) )
  {
    TCHAR   szOut[ 256 ] = { 0 };

    if( *szBuffer != '\0' )
    {
      _stprintf( szOut, TEXT( "Top-level window: %s" ), szBuffer );
      DEBUG_OUT( szOut );
    }
    
    if( !_tcsncmp( szBuffer,
                   TEXT( "Reggie" ),
                   _tcslen( TEXT( "Reggie" ) ) ) )
    {
      //
      //  The window has the correct title.  Enumerate through the child
      //    windows looking for a static text control containing the correct
      //    text.
      //
      //if( !EnumChildWindows( hWnd, EnumChildProc, (LPARAM)&bTextFound ) )
      //{
        //
        //  Add structured error handling.
        //
       //rc = FALSE;
      //}


      //if( TRUE == bTextFound )
      //{
        _stprintf( szBuffer, TEXT( "Found the window! [handle:%8.8x]" ), hWnd );
        DEBUG_OUT( szBuffer );
        //
        //  We have found the window we were looking for.  Return it to the
        //  thread procedure and exit the enumeration.
        //
        * ( HWND * ) lParam = hWnd;
        rc = FALSE;
      //}
    }
  }
  
EXIT:
  return rc;
}




//+++**************************************************************************
//  EnumChildProc()
//
//  DESCRIPTION:
//    A WNDENUMPROC-type function.  Called for each child window of the top-
//      level window being processed by EnumWindowsProc().  Checks for a static
//      text control containing the string "This is a test of the MBI service".
//
//  PARAMS:
//    HWND    hWnd    - handle to current child window
//    LPARAM  lParam  - points to a BOOL to recieve whether or not the static
//                      text control was found within the parent window
//
//  RETURNS:
//    FALSE (stops enumeration) on fatal error or if the window is found
//    TRUE (continues enumeration) if the supplied handle is not the window
//      we are looking for
//
//---**************************************************************************
BOOL
CALLBACK
EnumChildProc(
  HWND    hWnd,
  LPARAM  lParam 
  )
{
  BOOL    rc = TRUE;
  TCHAR   szBuffer[ 256 ] = { 0 };
  PBOOL   pbTextFound = (PBOOL)lParam;
  TCHAR   szOut[ 256 ] = { 0 };

  if( !pbTextFound )
  {
    rc = FALSE;
    goto EXIT;
  }

  SendMessage( hWnd, 
               WM_GETTEXT, 
               sizeof( szBuffer ) / sizeof( TCHAR ), 
               (LPARAM)&szBuffer );

  if( *szBuffer != '\0' )
  {
    _stprintf( szOut, TEXT( "Child window: %s" ), szBuffer );
    DEBUG_OUT( szOut );
  }
  
  if( _tcsstr( szBuffer,
               TEXT( "This is a test of the MBI service" ) ) )
  {
    //
    //  It does, so notify the EnumWindowsProc and exit the child
    //    enumeration.
    //
    *pbTextFound = TRUE;
    rc = FALSE;
  }

EXIT:
  return rc;
}


//+++**************************************************************************
//  DoApplicationSpecificHandling()
//
//  DESCRIPTION:
//    Performs application specific handling for the window in question, when
//      found.  For the purpose of this sample, it simply closes the window.
//
//  PARAMS:
//    HWND    hTargetWindow    - handle to window to process
//
//  RETURNS:
//    VOID
//
//---**************************************************************************
VOID
DoApplicationSpecificHandling(
  HWND    hTargetWindow 
  )
{
  //
  //  Here, we could do anything necessary... reboot the system, start a
  //    MAPI session and send an email, page a technician, etc.
  //

  SendMessage( hTargetWindow, WM_CLOSE, 0, (LPARAM)NULL );
}




///////////////////////////////////////////////////////////////////
//
//  The following code handles service installation and removal
//
//
//  FUNCTION: CmdInstallService()
//
//  PURPOSE: Installs the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdInstallService()
{
   SC_HANDLE   schService;
   SC_HANDLE   schSCManager;

   TCHAR szPath[512];
   TCHAR szErr[256];

   if ( GetModuleFileName( NULL, szPath, 512 ) == 0 )
   {
      _tprintf(TEXT("Unable to install %s - %s\n"), TEXT(SZSERVICEDISPLAYNAME), GetLastErrorText(szErr, 256));
      return;
   }

   schSCManager = OpenSCManager(
                               NULL,                   // machine (NULL == local)
                               NULL,                   // database (NULL == default)
                               SC_MANAGER_ALL_ACCESS   // access required
                               );
   if ( schSCManager )
   {
      schService = CreateService(
                                schSCManager,               // SCManager database
                                TEXT(SZSERVICENAME),        // name of service
                                TEXT(SZSERVICEDISPLAYNAME), // name to display
                                SERVICE_ALL_ACCESS,         // desired access
                                SERVICE_INTERACTIVE_PROCESS | SERVICE_WIN32_OWN_PROCESS,  // service type
                                SERVICE_AUTO_START,       // start type
                                SERVICE_ERROR_NORMAL,       // error control type
                                szPath,                     // service's binary
                                NULL,                       // no load ordering group
                                NULL,                       // no tag identifier
                                TEXT(SZDEPENDENCIES),       // dependencies
                                NULL,                       // LocalSystem account
                                NULL);                      // no password

      if ( schService )
      {
         _tprintf(TEXT("%s installed.\n"), TEXT(SZSERVICEDISPLAYNAME) );
         CloseServiceHandle(schService);
      }
      else
      {
         _tprintf(TEXT("CreateService failed - %s\n"), GetLastErrorText(szErr, 256));
      }

      CloseServiceHandle(schSCManager);
   }
   else
      _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));
}



//
//  FUNCTION: CmdRemoveService()
//
//  PURPOSE: Stops and removes the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdRemoveService()
{
  
   
   SC_HANDLE   schService;
   SC_HANDLE   schSCManager;

   TCHAR       szErr[256];

   schSCManager = OpenSCManager(
                               NULL,                   // machine (NULL == local)
                               NULL,                   // database (NULL == default)
                               SC_MANAGER_ALL_ACCESS   // access required
                              ); 

   if ( schSCManager )
   {
      schService = OpenService(schSCManager, TEXT(SZSERVICENAME), SERVICE_ALL_ACCESS);

      if (schService)
      {
         // try to stop the service
         if ( ControlService( schService, SERVICE_CONTROL_STOP, &g_svcStatus ) )
         {
            _tprintf(TEXT("Stopping %s."), TEXT(SZSERVICEDISPLAYNAME));
            Sleep( 1000 );

            while ( QueryServiceStatus( schService, &g_svcStatus) )
            {
               if ( g_svcStatus.dwCurrentState == SERVICE_STOP_PENDING )
               {
                  _tprintf(TEXT("."));
                  Sleep( 1000 );
               }
               else
                  break;
            }

            if ( g_svcStatus.dwCurrentState == SERVICE_STOPPED )
               _tprintf(TEXT("\n%s stopped.\n"), TEXT(SZSERVICEDISPLAYNAME) );
            else
               _tprintf(TEXT("\n%s failed to stop.\n"), TEXT(SZSERVICEDISPLAYNAME) );

         }

         // now remove the service
         if ( DeleteService(schService) )
            _tprintf(TEXT("%s removed.\n"), TEXT(SZSERVICEDISPLAYNAME) );
         else
            _tprintf(TEXT("DeleteService failed - %s\n"), GetLastErrorText(szErr,256));


         CloseServiceHandle(schService);
      }
      else
         _tprintf(TEXT("OpenService failed - %s\n"), GetLastErrorText(szErr,256));

      CloseServiceHandle(schSCManager);
   }
   else
      _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText(szErr,256));
}



//
//  FUNCTION: GetLastErrorText
//
//  PURPOSE: copies error message text to string
//
//  PARAMETERS:
//    lpszBuf - destination buffer
//    dwSize - size of buffer
//
//  RETURN VALUE:
//    destination buffer
//
//  COMMENTS:
//
LPTSTR GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize )
{
   DWORD dwRet;
   LPTSTR lpszTemp = NULL;

   dwRet = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_ARGUMENT_ARRAY,
                          NULL,
                          GetLastError(),
                          LANG_NEUTRAL,
                          (LPTSTR)&lpszTemp,
                          0,
                          NULL );

   // supplied buffer is not long enough
   if ( !dwRet || ( (long)dwSize < (long)dwRet+14 ) )
      lpszBuf[0] = TEXT('\0');
   else
   {
      lpszTemp[lstrlen(lpszTemp)-2] = TEXT('\0');  //remove cr and newline character
      _stprintf( lpszBuf, TEXT("%s (0x%x)"), lpszTemp, GetLastError() );
   }

   if ( lpszTemp )
      LocalFree((HLOCAL) lpszTemp );

   return lpszBuf;
}
